def assertEqual(a, b):
    assert a == b


def assertNotEqual(a, b):
    assert a != b


def assertTrue(a):
    assert a is True


def assertFalse(a):
    assert a is False


def assertLessEqual(a, b):
    assert a <= b


def assertLess(a, b):
    assert a < b


def assertGreaterEqual(a, b):
    assert a >= b


def assertGreater(a, b):
    assert a > b


def assertIn(a, b):
    assert a in b


def assertNotIn(a, b):
    assert a not in b


def assertIsNot(a, b):
    assert a is not b
